ClickCommander = {}
ClickCommander.version = '1.0'

--## Choco ##--

local cc = ClickCommander
cc.vars = dofile'vars.lua'

local aliases --(table) keep track of which aliases exist {[name1_lower]={name1_orig_case,alias_str}...}
local is_intercept_chat = false --enable chatreceiver when true
local last_command --keep track of previous command from command line

--characters not allowed for user-defined command names
local illegal_characters = {
	['%s'] = "any spaces",
	[','] = "any commas",
	['[{}]'] = "curly brackets {}",
	['"'] = "double quotes",
	["'"] = "single quotes",
	[':'] = "the colon character",
}

--vars settings
local all_vars = cc.vars.GetAll()
local commands = {}
local HOLD_DELAY = all_vars.hold_delay
local TAP_DELAY = all_vars.tap_delay


--// intercepts alias list from chat
local chatreceiver_old = chatreceiver.OnEvent
function chatreceiver:OnEvent(event, data, ...)
	if is_intercept_chat and event=="CHAT_MSG_PRINT" then
		local cmd_name, cmd_str = data.msg:match"^(%S+) = (.+)$"
		if cmd_name and cmd_str then
			aliases[cmd_name:lower()] = {cmd_name, cmd_str} --keep track of aliases that are defined
			
			return --hide this line from chat
		end
	end
	chatreceiver_old(chatreceiver, event, data, ...)
end


--// Prints message to chat with custom coloring and prefixed with plugin name
	--to print a single line of text, str can be a string
	--to print multiple lines of text, use a table for str instead
local function cc_print(str)
	local lines
	if type(str)~='table' then
		lines = {tostring(str)}
	else lines = str end
	
	for i,line_entry in ipairs(lines) do
		local line = tostring(line_entry)
		local substr = line:match"^Error: (.*)"
		
		line = substr or line
		local error_prefix = substr and " Error" or ''
		local title = i==1 and '\127ffffffClickCommander'..error_prefix..': \127a0a0a0' or '\t'
		local closing = i==#lines and '\127o\127o' or ''
		
		print(title..line..closing)
	end
end


--// Starts a timer for the minimum amount of time to hold button down to be a hold click
local function start_hold_timer(cmd)
	local info = commands[cmd]
	if not info then return end
	
	if info.timer:IsActive() then info.timer:Kill() end
	info.timer:SetTimeout(HOLD_DELAY, function()
		if info.mode=="hold" then
			local alias = cmd.."_hold"
			if aliases[alias] then gkinterface.GKProcessCommand(alias) end
			
			--reset
			info.mode = nil
			info.tap_count = 0
		end
	end)
end


--// Starts a timer for the maximum amount of time between clicks to be considered a double-click
local function start_tap_timer(cmd)
	local info = commands[cmd]
	if not info then return end
	
	local alias = cmd.."_tap"
	if aliases[alias] then gkinterface.GKProcessCommand(alias) end
	
	if info.timer:IsActive() then info.timer:Kill() end
	info.timer:SetTimeout(TAP_DELAY, function()
		alias = cmd.."_tap"..info.tap_count
		if aliases[alias] then gkinterface.GKProcessCommand(alias) end
		
		--reset
		info.mode = nil
		info.tap_count = 0
	end)
end


--// Called when the monitored button is pressed or released. Determines which function to execute depending on type of click
local function button_action(cmd, is_down)
	local info = commands[cmd]
	if not info then return end
	
	if not aliases then cc:refresh{'silent'} end --get current aliases on first run only
	
	if is_down and not info.is_active and not info.mode then --new click, start hold timer
		info.mode = "hold"
		info.tap_count = 0
		start_hold_timer(cmd)
	elseif not is_down and info.is_active and info.mode=="hold" then --first fast tap
		info.mode = "click"
		info.tap_count = 1
		start_tap_timer(cmd)
	elseif is_down and not info.is_active and info.mode=="click" then --additional fast tap(s)
		info.tap_count = info.tap_count + 1
		start_tap_timer(cmd)
	elseif not is_down and info.is_active and not info.mode then --reset after hold done
		info.tap_count = 0
	end --ignore button release in modes "click" or nil
	
	info.is_active = is_down --ignore repeated button presses after first (until button released again)
end


--// Writes list of commands to config.ini
local function save_commands()
	local command_list = {}
	for cmd,info in pairs(commands) do
		local cmd_case = info.name
		table.insert(command_list, cmd_case)
	end
	table.sort(command_list)
	
	cc.vars.SaveVars('commands', command_list) 
end


--// Adds a new command to the player-defined list of commands
	--cmd (string) - name of the command to add (not case-sensitive, but case preserved)
	--returns true if succes, otherwise false with error string
local function add_command(cmd)
	assert(type(cmd)=="string", "Bad argument #1 to 'add_command' (string expected)")
	if cmd=='' then return false, "command name must contain at least one character" end
	
	local cmd_lower = cmd:lower()
	
	if not commands[cmd_lower] then
		--check for illegal characters
		for pattern,desc in pairs(illegal_characters) do
			if cmd:match(pattern) then return false, "command name must not contain "..desc end
		end
		
		commands[cmd_lower] = {
			name = cmd, --preserve case
			timer = Timer(),
			tap_count = 0,
		}
		
		RegisterUserCommand("+"..cmd, function() button_action(cmd_lower, true) end)
		RegisterUserCommand("-"..cmd, function() button_action(cmd_lower, false) end)
		
		return true
	else return false, "the command '"..cmd.."' already exists" end
end


--// verifies new delay value is valid
	--returns true and value if valid, or returns false and error string
local function set_delay(key, val)
	assert(all_vars[key], "Bad argument #1 to 'set_delay', not a valid key name")
	local delay = tonumber(val)
	
	local str = type(val)=="string" and val:lower()
	if str=="default" then delay = cc.vars.GetDefault(key) end
	
	if delay then
		if delay <= 2 then
			delay = delay*1000 --convert to milliseconds
		elseif delay > 2000 then return false,'delay cannot be more than 2 seconds' end
		
		if delay < 1 then return false,'delay must be at least 1 ms' end
		
		cc.vars.SaveVars(key, delay)
		return true, delay
	else return false,"delay must be a number; use 'default' to restore default value" end
end


--// Processes command-line inputs
local inputs = {add=true,delete=true,list=true,delay=true,refresh=true,help=true}
function cc:input(args)
	args = args or {}
	assert(type(args)=="table", "Bad arguments to 'input' (must be formatted as table)")
	
	local command = args[1] and args[1]:lower()
	
	if inputs[command] then
		local this_command = table.concat(args, ' ')
		
		table.remove(args, 1)
		cc[command](self, args)
		
		last_command = this_command
	else cc.help(self) end --default command
end


--// Adds a new command named by the player
function cc:add(args)
	args = args or {}
	assert(type(args)=='table', "Bad arguments to 'add' (must be formatted as table)")
	local arg1 = table.remove(args, 1)
	assert(not arg1 or type(arg1)=="string", "Bad argument #1 in 'add' (string expected)")
	
	if arg1 then
		local is_success,msg = add_command(arg1)
		if is_success then
			save_commands() --update config.ini
			cc_print(arg1.." command created")
		else cc_print("Error: "..msg) end
	else cc.help(self, {"add"}) end
end


--// Removes existing commands setup by the player
function cc:delete(args)
	args = args or {}
	assert(type(args)=='table', "Bad arguments to 'delete' (must be formatted as table)")
	local arg1 = table.remove(args, 1)
	assert(not arg1 or type(arg1)=="string", "Bad argument #1 in 'delete' (string expected)")
	
	if arg1 then
		local arg1_lower = arg1:lower()
		
		if commands[arg1_lower] then
			commands[arg1_lower] = nil
			
			UnregisterUserCommand("+"..arg1)
			UnregisterUserCommand("-"..arg1)
			
			save_commands() --update config.ini
			
			cc_print(arg1.." command deleted")
		else cc_print("Error: command '"..arg1.."' does not exist") end
	else cc.help(self, {"delete"}) end
end


--// Prints a list of commands the player has setup
function cc:list(args)
	args = args or {}
	assert(type(args)=='table', "Bad arguments to 'list' (must be formatted as table)")
	local arg1 = table.remove(args, 1)
	assert(not arg1 or type(arg1)=="string", "Bad argument #1 in 'list' (string expected)")
	
	if arg1 then --print list of aliases associated with arg1 command
		local arg1_lower = arg1:lower()
		
		if commands[arg1_lower] then
			if not aliases then cc:refresh{'silent'} end --get current aliases on first run only
			
			local list = {} --list of aliases corresponding to arg1
			for cmd_lower,info in pairs(aliases) do
				local cmd_name,cmd_str = unpack(info)
				
				local sort_order = -1
				local prefix,suffix = cmd_lower:match"^(%S+)(_hold)$"
				if not suffix then
					prefix,suffix = cmd_lower:match"^(%S+)(_tap)$"
					sort_order = 0
				end
				if not suffix then
					prefix,suffix = cmd_lower:match"^(%S+)_tap(%d+)$"
					sort_order = tonumber(suffix)
				end
				
				if prefix and prefix:lower()==arg1_lower then
					table.insert(list, {sort_order, cmd_name.." = "..cmd_str})
				end
				
			end
			table.sort(list, function(a, b) return a[1] < b[1] end)
			
			local lines = {"Aliases for '"..arg1.."':"..(#list>0 and "" or " none")}
			for _,item in ipairs(list) do
				table.insert(lines, item[2])
			end
			
			cc_print(lines)
		else cc_print("command '"..arg1.."' has not been defined") end
	else -- print list of user-defined commands
		local list = {}
		for cmd,info in pairs(commands) do
			local cmd_case = info.name
			table.insert(list, cmd_case)
		end
		table.sort(list)
	
		if #list>0 then
			cc_print("Active commands: "..table.concat(list, ", "))
		else cc_print"No commands created" end
	end
end


--// Updates list of aliases - this function needs to be called after changing aliases
	--if args[1]=="silent" then don't include print statement
function cc:refresh(args)
	args = args or {}
	assert(type(args)=='table', "Bad argument #1 to 'refresh' (table expected)")
	
	local arg1 = table.remove(args, 1)
	if type(arg1)=='string' then arg1 = arg1:lower() end
	
	--update aliases
	aliases = {} --clear previous data
	is_intercept_chat = true --enable chatreceiver interceptor
	gkinterface.GKProcessCommand"alias"
	--console_clear() --TODO this doesn't have any affect until reloading interface; need better solution
	is_intercept_chat = false
	
	if arg1~='silent' then cc_print"updated based on current aliases" end
end


--// Changes the delay time settings
function cc:delay(args)
	args = args or {}
	assert(type(args)=='table', "Bad arguments to 'delay' (must be formatted as table)")
	
	local arg1 = table.remove(args, 1)
	assert(not arg1 or type(arg1)=="string", "Bad argument #1 in 'delay' (string expected)")
	local arg1_lower = arg1 and arg1:lower()
	
	local delay = table.remove(args, 1) --arg2
	assert(not delay or type(delay)=="string" or type(delay)=="number", "Bad argument #2 in 'delay' (string or number expected)")
	
	if arg1_lower=="hold" then
		if delay then
			local is_success,msg = set_delay('hold_delay', delay)
			if is_success then
				HOLD_DELAY = msg
				cc_print("Max hold delay changed to "..msg.."ms")
			else cc_print(msg) end --print error message
		else cc_print("Max hold delay currently set to: "..HOLD_DELAY.."ms") end --print current value to chat
	elseif arg1_lower=="tap" then
		if delay then
			local is_success,msg = set_delay('tap_delay', delay)
			if is_success then
				TAP_DELAY = msg
				cc_print("Max tap delay changed to "..msg.."ms")
			else cc_print(msg) end --print error message
		else cc_print("Max delay between multi-taps currently set to: "..TAP_DELAY.."ms") end --print current value to chat
	else cc.help(self, {"delay"}) end
end


--// Prints help text
function cc:help(args)
	args = args or {}
	assert(type(args)=='table', "Bad arguments to 'help' (must be formatted as table)")
	
	local help_topics = {'add','delete','list','delay','refresh','usage'} --order to display help topics
	
	local help_text = {
		['TITLE'] = "\127ffffffClickCommander v"..cc.version.."  Available commands are: (for more info use /cc help \127ffff00COMMAND\127o)",
		['add'] = {"add \127ffff00ALIAS_PREFIX\127o", "Adds a new command (user-defined) to monitor for button presses", ". For more info see \127ffff00/cc help usage\127o",
			false, "The alias prefix name should not match any existing commands or aliases and must not contain any spaces",
			false, "You'll want to bind the alias prefix to a keyboard or joystick button by using the following: \127ffff00/bind BUTTON_NAME +ALIAS_PREFIX\127o",
		},
		['delete'] = {"delete \127ffff00alias_prefix\127o", "Removes an existing command and unbinds any associated buttons", false,
			false, "To see a list of current commands, use \127ffff00/cc list\127o",
		},
		['list'] = {"list", "Lists the currently active commands", false,
			"list \127ffff00ALIAS_PREFIX\127o", "Prints the buttons bound to this command",
		},
		['delay'] = {"delay hold|tap \1270066ffTIME\127o", "Changes the delay time for a long press or double-click, in milliseconds", false,
			false, "The word default can be used instead of specifying a time to return to default settings: e.g. \127ffff00/cc delay hold default\127o",
			false, "Omitting the time value will print the current setting instead of changing it, e.g. \127ffff00/cc delay tap\127o",
		},
		['refresh'] = {"refresh", "Run this command after changing aliases", false,
		},
		['usage'] = {"help usage", "Describes how to bind buttons and setup the associated aliases needed to use this plug-in", false,
			false, "First setup a new command using \127ffff00/cc add my_command\127o",
			false, "Second, bind the newly created command to a keyboard or joystick button: e.g. \127ffff00/bind MMBUTTON +my_command\127o",
			false, "Finally, create aliases to define the behavior of each type of button press: (for examples do \127ffff00/cc help example\127o)",
			false, "\tUse an alias ending in _hold for a long press",
			false, "\tFor a single tap, double tap, etc. use _tap1, _tap2, _tap3, etc.",
			false, "\tNote that the alias ending in _tap will also get executed for all non-hold presses (if the alias is defined)",
		},
		['example'] = {"help example", "Example usage:", false,
			"add my_command", "create a new command with a user-defined name",
			"/bind MMBUTTON +my_command", "bind the command to the middle mousebutton",
			"/alias my_command_hold FlyModeToggle", "When button is held down the flight mode is changed",
			"/alias my_command_tap1 RadarNextNearestEnemy", "When button is quickly pressed and released, the nearest enemy is targeted",
			"/alias my_command_tap2 RadarNextFront", "When the button is double-clicked, the object in front is targeted",
			"refresh", "Run this command after changing your aliases to re-sync",
		},
	}
	
	local cmd = args[1] and args[1]:lower()
	
	local help_lines = {}
	local cmd_text = help_text[cmd]
	
	if not cmd_text then --default prints top-level help
		table.insert(help_lines, help_text.TITLE) --introductory line with version
		
		for _,topic in ipairs(help_topics) do
			if help_text[topic] then
				local command = help_text[topic][1]
				local desc = help_text[topic][2]
				local extra = help_text[topic][3] or "" --extra text to only display at top-level
				
				local cmd_str = ''
				if command then
					local is_ext_cmd = command:match"^/"
					cmd_str = (is_ext_cmd and "" or "/cc ")..command.." - "
				end
				table.insert(help_lines, "\t"..cmd_str..desc..extra)
			end
		end
	else --command-specific help
		table.insert(help_lines, "\127ffffff/cc "..cmd_text[1].." - "..cmd_text[2])
		
		for i=4,#cmd_text,2 do
			local command = cmd_text[i]
			local desc = cmd_text[i+1]
			
			local cmd_str = ''
			if command then
				local is_ext_cmd = command:match"^/"
				cmd_str = (is_ext_cmd and "" or "/cc ")..command.." - "
			end
			table.insert(help_lines, "\t"..cmd_str..desc)
		end
	end
	
	print(table.concat(help_lines, "\n"))
end


--initialize commands data from config.ini
for _,cmd in ipairs(all_vars.commands or {}) do
	add_command(cmd) --includes registering user commands
end

RegisterUserCommand("CC", cc.input)